// Lua interface
// Copyright (c) 2012 by Kasji

#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include "lua.h"
#ifdef __cplusplus
#include <lua5.3/lua.hpp>
#else
//extern "C" {
#include <lua5.3/lua.h>
#include <lua5.3/lualib.h>
#include <lua5.3/lauxlib.h>
//}
#endif // __cplusplus

lua_State * g_luaState;
L_UPDATER * g_luaUpdater;
jmp_buf g_luaEnv;

// Global funs
DECLARE_LUA_FUN(L_send_to_char);
DECLARE_LUA_FUN(L_get_char_world);
DECLARE_LUA_FUN(L_get_char_room);
DECLARE_LUA_FUN(L_get_room_index);
DECLARE_LUA_FUN(L_interpret);
DECLARE_LUA_FUN(L_open_passage);
DECLARE_LUA_FUN(L_close_passage);
DECLARE_LUA_FUN(L_say);
DECLARE_LUA_FUN(L_recho);
DECLARE_LUA_FUN(L_rechoaround);
DECLARE_LUA_FUN(L_mudtime);

// Getter/Setter funs
DECLARE_LUA_FUN(luaMudGet);
DECLARE_LUA_FUN(luaCHGet);
DECLARE_LUA_FUN(luaCHSet);
DECLARE_LUA_FUN(luaRoomGet);
DECLARE_LUA_FUN(luaRoomSet);
DECLARE_LUA_FUN(luaObjGet);
DECLARE_LUA_FUN(luaObjSet);

static const luaL_Reg luaMudlib[] =
{
	{"send_to_char", L_send_to_char},
	{"get_char_room", L_get_char_room},
	{"get_char_world", L_get_char_world},
	{"get_room_index", L_get_room_index},
	{"interpret", L_interpret},
	{"open_passage", L_open_passage},
	{"close_passage", L_close_passage},
	{"recho", L_recho},
	{"rechoaround", L_rechoaround},
	{NULL, NULL}
};

void GetTracebackFunction (lua_State *L)
{
   lua_pushliteral (L, LUA_DBLIBNAME);     /* "debug"   */
   lua_rawget      (L, LUA_RIDX_GLOBALS);    /* get debug library   */

   if (!lua_istable (L, -1))
   {
      lua_pop (L, 2);   /* pop result and debug table  */
      lua_pushnil (L);
      return;
   }

   /* get debug.traceback  */
   lua_pushstring(L, "traceback");
   lua_rawget    (L, -2);               /* get traceback function  */

   if (!lua_isfunction (L, -1))
   {
      lua_pop (L, 2);   /* pop result and debug table  */
      lua_pushnil (L);
      return;
   }

   lua_remove (L, -2);   /* remove debug table, leave traceback function  */
}  /* end of GetTracebackFunction */

int CallLuaWithTraceBack (lua_State *L, const int iArguments, const int iReturn)
{
   int error;
   int base = lua_gettop (L) - iArguments;  /* function index */
   GetTracebackFunction (L);
   if (lua_isnil (L, -1))
   {
      lua_pop (L, 1);   /* pop non-existent function  */
      error = lua_pcall (L, iArguments, iReturn, 0);
   }
   else
   {
      lua_insert (L, base);  /* put it under chunk and args */
      error = lua_pcall (L, iArguments, iReturn, base);
      lua_remove (L, base);  /* remove traceback function */
   }

   return error;
}  /* end of CallLuaWithTraceBack  */

extern "C" int luaPanic(lua_State * L)
{
	const char * msg = lua_tostring(L, -1);

	log_lua("FATAL ERROR: %s", msg);
	longjmp(g_luaEnv, 1);

	return 0; // Never returns
}

extern "C" int luaRegisterMudlib(lua_State * L)
{
	int i = 0;

	while (luaMudlib[i].name != NULL && luaMudlib[i].func != NULL)
	{
		lua_register(L, luaMudlib[i].name, luaMudlib[i].func);
		i++;
	}
	return RET_OK;
}

int luaRegisterEnviron(lua_State * L)
{
   // Setup the environment.
   lua_newtable(L);
   lua_replace(L, LUA_RIDX_MAINTHREAD);

   return 0;
}

int luaOpen()
{
	// Create a new Lua state.
	g_luaState = luaL_newstate();
	if (!g_luaState)
	{
		log_lua("luaOpen: Failed to create lua_State!");
		return RET_ERR;
	}
//	lua_gc(g_luaState, LUA_GCRESTART, 0);
	// Error recovery.
	lua_atpanic(g_luaState, luaPanic);
	// Load standard libraries.
	luaL_openlibs(g_luaState);
	// Load mud library.
	luaRegisterMudlib(g_luaState);
	// Set package.path
	lua_getglobal(g_luaState, "package");
	lua_pushstring(g_luaState, LUA_DIR "?.lua");
	lua_setfield(g_luaState, -2, "path");
	// Empty the stack.
	lua_settop(g_luaState, 0);

	// Start the coroutine updater.
	g_luaUpdater = new L_UPDATER;

	return RET_OK;
}

int luaClose()
{
	lua_close(g_luaState);

	return RET_OK;
}

void log_lua(const char * str, ...)
{
   char buf[MAX_STRING_LENGTH];

//   strcpy( buf, "[*****] BUG: " );
   {
      va_list param;

      va_start( param, str );
      vsprintf( buf /*+ strlen( buf )*/, str, param );
      va_end( param );
   }
   log_string_plus(buf, LOG_LUA, LEVEL_IMMORTAL);

   return;
}

void lua_command(CHAR_DATA * ch, const char * command, const char * argument)
{
	L_COROUTINE * thread;
   char path[MAX_INPUT_LENGTH];
   char func[MAX_INPUT_LENGTH];
   int err;
	int jmp;
	int result;

	jmp = setjmp(g_luaEnv);
	if (jmp == 1)
	{
		log_lua("Long jump successful.");
		return;
	}

   if (!ch)
   {
      log_lua("lua_command: FATAL error: ch == NULL!");
      return;
   }

   if (!g_luaState)
   {
      log_lua("lua_command: FATAL error: g_luaState == NULL!");
      return;
   }

	// Thread it.
	thread = g_luaUpdater->newthread("char", ch->GetId().Value());

   // Clear the environment.
   lua_pushcfunction(thread->L(), luaRegisterEnviron);
   lua_call(thread->L(), 0, 0);
	lua_registerMud(thread->L());
	lua_settop(thread->L(), 0);

	// mud.lua
	snprintf(path, MAX_INPUT_LENGTH, "%smud.lua", LUA_DIR);
	luaL_dofile(thread->L(), path);

   lua_pushlightuserdata(thread->L(), (void*)ch->GetId().Value());
   lua_setglobal(thread->L(), CHAR_ID);

   snprintf(func, MAX_INPUT_LENGTH, "do_%s", command);
/*   lua_getglobal(thread, func);
   if (!lua_isfunction(thread, -1))
   {
      log_lua("lua_command: Function not found: %s", func);
//      lua_settop(g_luaState, 0);
      return;
   } */

   // Setup the ch metatable.
	lua_pushCH(thread->L(), ch);

	// do_?.lua
   snprintf(path, MAX_INPUT_LENGTH, "%sdo_%s.lua", LUA_CMD_DIR, command);
   if ((err = luaL_dofile(thread->L(), path)) != 0)
   {
      log_lua("lua_command: luaL_dofile() error %d: %s\n\r   Path: %s", err, lua_tostring(g_luaState, -1), path);
      return;
   }

   if (!argument || argument[0] == '\0')
      lua_pushnil(thread->L());
   else
      lua_pushstring(thread->L(), argument);
//   if ((err = lua_pcall(g_luaState, 2, 0, 0)) != 0)

	result = thread->run(func, 2);
	if (result != LUA_OK && result != LUA_YIELD)
   {
		const char * msg;
		log_lua("lua_command: thread->run() failed.");
		if (lua_isstring(thread->L(), -1))
		{
			log_lua("%s", lua_tostring(thread->L(), -1));
			lua_pop(thread->L(), 1);
		}
//      log_lua("lua_command: lua_pcall() error: %d: %s\n\r   Path: %s", err, lua_tostring(g_luaState, -1), path);
//      lua_settop(g_luaState, 0);
      return;
   }
//   lua_settop(g_luaState, 0);

   return;
}

void do_lua(CHAR_DATA * ch, const char * argument)
{
    char arg1[MAX_INPUT_LENGTH];
//    char arg2[MAX_INPUT_LENGTH];

    argument = one_argument(argument, arg1);

    if (arg1[0] == '\0')
    {
        send_to_char("Syntax: lua <command> <args>\n\r", ch);
        send_to_char("Available Commands:\n\r", ch);
        send_to_char("  dir    cd    mkdir    edit    test\n\r", ch);
        return;
    }

    if (!str_cmp(arg1, "dir"))
    {
    }

    return;
}

void lua_registerMud(lua_State * L)
{
   if (!L)
      return;

   // Setup the mud metatable.
   lua_newtable(L);
	lua_pushcfunction(L, L_mudtime);
	lua_setfield(L, -2, "time");
   lua_newtable(L);
   lua_pushcfunction(L, luaMudGet);
   lua_setfield(L, -2, "__index");
   lua_setmetatable(L, -2);
	lua_setglobal(L, "mud");

   return;
}

void lua_pushCH(lua_State * L, CHAR_DATA * ch)
{
	if (!L)
		return;
	if (!ch)
	{
		lua_pushnil(L);
		return;
	}

   // Setup the ch metatable.
   lua_newtable(L);
	lua_pushstring(L, "char");
	lua_setfield(L, -2, "type");
   lua_pushlightuserdata(L, (void*)ch->GetId().Value());
   lua_setfield(L, -2, "id");
	lua_pushcfunction(L, L_say);
	lua_setfield(L, -2, "say");
   lua_newtable(L);
   lua_pushcfunction(L, luaCHGet);
   lua_setfield(L, -2, "__index");
   lua_setmetatable(L, -2);

	return;
}

void lua_pushObj(lua_State * L, OBJ_DATA * obj)
{
	if (!L)
		return;
	if (!obj)
	{
		lua_pushnil(L);
		return;
	}

	// Setup the obj metatable
	lua_newtable(L);
	lua_pushstring(L, "obj");
	lua_setfield(L, -2, "type");
	lua_pushlightuserdata(L, (void*)obj->GetId().Value());
	lua_setfield(L, -2, "id");
	lua_newtable(L);
	lua_pushcfunction(L, luaObjGet);
	lua_setfield(L, -2, "__index");
	lua_setmetatable(L, -2);

	return;
}

void lua_pushRoom(lua_State * L, ROOM_INDEX_DATA * room)
{
   if (!L)
      return;
   if (!room)
   {
      lua_pushnil(L);
      return;
   }

   // Setup the room metatable
   lua_newtable(L);
	lua_pushstring(L, "room");
	lua_setfield(L, -2, "type");
   lua_pushlightuserdata(L, (void*)room->GetId().Value());
   lua_setfield(L, -2, "id");
   lua_newtable(L);
   lua_pushcfunction(L, luaRoomGet);
   lua_setfield(L, -2, "__index");
   lua_setmetatable(L, -2);

   return;
}

char * lua_check_filename(char * filename)
{
	char * it = filename;

	if (!filename)
		return NULL;

	while ((*it) != '\0')
	{
		if (((*it) >= '0' && (*it) <= '9') || ((*it) >= 'a' && (*it) <= 'z')
			 || ((*it) >= 'A' && (*it) <= 'Z') || (*it) == '_' || (*it) == '-')
		{
			it++;
			continue;
		}
		if ((*it) == '\n' || (*it) == '\r')
		{
			*it = '\0';
			break;
		}
		*it = '_';
		it++;
	}

	return filename;
}

void do_luamem(CHAR_DATA * ch, const char * argument)
{
	if (g_luaState)
		ch_printf(ch, "Lua Memory: %dKb\n\r", lua_gc(g_luaState, LUA_GCCOUNT, 0));
	else
		send_to_char("Lua not ready.\n\r", ch);
	return;
}
